/*
 * AList.e
 *
 * AList Mailing List Main Server.
 *
 * This program and all associated sources and documents are protected under
 *   the GNU Public License, version 2 or higher.
 *
 * History:    (See history log for details)
 *    Oct 30, 1995   jlm   thought@weblink.org, Created (v1.0)
 *    Nov 13, 1995   jlm   Version 0.2 released.
 *    Nov 16, 1995   jlm   Version 0.3
 *    Oct 16, 1997   jlm   Version 0.4.  Added AlternateCmd in AList.conf.
 */

OPT LARGE

MODULE 'dos/var'
MODULE 'dos/dos'
MODULE 'exec/memory'
MODULE 'exec/nodes'
MODULE 'exec/ports'
MODULE 'other/aldef'
MODULE 'other/alconfig'
MODULE 'other/aladdr'
MODULE 'other/allog'
MODULE 'other/alpost'
MODULE 'other/alcmd'
MODULE 'other/aladd'

/*
 * list_dir is a lock on the directory for the lists' faqs, member lists, and config files.
 *
 * files_dir is a lock on the directory with sub-directories for each list with files and/or archives.
 *
 * config is a lock on the global alink config file.  Only non-NIL when in use.
 */
DEF list_dir=NIL, files_dir=NIL, config=NIL, myport=NIL:PTR TO mp, password=NIL:PTR TO CHAR, no_from=NIL
DEF config_list=NIL:PTR TO config_node, temp_file=NIL:PTR TO CHAR, log_dir=NIL, hostname=NIL:PTR TO CHAR
DEF alternate_cmd=NIL:PTR TO CHAR

PROC main()
   DEF tmp

   '$VER:AList Mailing List Server, version 0.4, compiled on October 16, 1997\n'

   log_dir := Lock ('', ACCESS_READ)

   /* This should always be the first thing we do, except assigning the log_dir */
   log_message ('Going up.\n', LOG_INFO);

   tmp := NIL
   IF (arg)
      config := Open (arg, MODE_OLDFILE)
   ENDIF
   IF (config = NIL)
      tmp := String (64)
      IF GetVar ('AList_Config', tmp, 64, GVF_GLOBAL_ONLY OR GVF_LOCAL_ONLY)
         config := Open (tmp, MODE_OLDFILE)
      ENDIF
   ENDIF
   IF (config = NIL)
      config := Open ('AList.conf', MODE_OLDFILE)
      IF (config = NIL)
         log_message ('Failure to locate config file.\n', LOG_ERROR)
         CleanUp (20)
      ENDIF
   ENDIF

   /* This had BETTER close the config file!  It also locks list_dir and files_dir. */
   /* It reads in all the list config files, too. */
   parse_config ()
   DisposeLink (tmp)

   IF (config_list = NIL)
      /* No point in going on if we're not serving any lists... */
      JUMP end_it_all
   ENDIF

   /* By the time we get here, we're ready for business! */
   myport := CreateMsgPort ()
   IF (myport = NIL)
      log_message ('Failure to create message port!\n', LOG_ERROR)
      CleanUp (20)
   ENDIF

   myport.ln.name := 'AList'
   myport.ln.pri := 1

   AddPort (myport)

   barnacle_start()

   main_loop()

   barnacle_end()

   /* Dump all our remaining messages, we're going down! */
   /* Since we already have this code, let's use it! */
   alist_replyport := myport
   barnacle_end()

end_it_all:
   /* Dump our config stuff */
   free_config ()


   /* This should always be the last thing we do, except freeing the log_dir. */
   log_message ('Going down.\n', LOG_INFO)

   IF (log_dir)  THEN UnLock (log_dir)
ENDPROC


/*
 * Handles and distributes the incoming messages and commands.
 */
PROC main_loop()
   DEF test_bit, done, b, msg:PTR TO almsg, buff, reply, str1:PTR TO CHAR, str2:PTR TO CHAR, str3

   test_bit := Shl (1, myport.sigbit)
   done := FALSE
   b := TRUE
   buff := String (256)
   WHILE (done = FALSE)
      IF (b = FALSE)  THEN Wait (test_bit)
      msg := GetMsg (myport)
      IF (msg)
         b := TRUE
         /* Use a jump table of size ALM_NOOP, the last command */
         SELECT ALM_NOOP OF msg.command
         CASE ALM_QUIT
            done := TRUE

         CASE ALM_SEND
            str2 := NIL
            str1 := NIL
            IF (msg.field1)   /* List name */
               str2 := String (StrLen (msg.field1))
               StrCopy (str2, msg.field1)
            ELSE
               str2 := String (1)
            ENDIF
            str1 := add_link (str1, str2)
            IF (msg.field2)
               str2 := String (StrLen (msg.field2))
               StrCopy (str2, msg.field2)
            ELSE
               str2 := String (1)
            ENDIF
            str1 := add_link (str1, str2)
            str3 := msg.arg
            WHILE (str3)
               str2 := String (EstrLen (str3))
               StrCopy (str2, str3)
               str1 := add_link (str1, str2)
               str3 := Next (str3)
            ENDWHILE
            b := ALM_SEND

         CASE ALM_ADDR
            /* Note that this truncates the From: and To: lines (which are parsed by this) to 255 chars */
            IF (msg.arg)
               StrCopy (buff, msg.arg)
            ELSE
               buff[0] := NIL
            ENDIF
            b := ALM_ADDR

         DEFAULT
            StringF (buff, 'Unknown command #\d received on message port.\n', msg.command)
            log_message (buff, LOG_WARNING)
         ENDSELECT

         reply := msg.mn.replyport
         /* Let our port start receiving again */
         ReplyMsg (msg)
         /* Note:  you'd BETTER be expecting it to NT_FREEMSG your message
                   if you didn't include a replyport! */

         /* Do something if we need to */
         IF (b <> TRUE)
            SELECT ALM_NOOP OF b
            CASE ALM_ADDR
               /* name, address */
               str1,str2 := parse_addr (buff)

               boat_put_msg (reply, - ALM_ADDR, NIL, str1, str2)

               DisposeLink (str1)
               DisposeLink (str2)

            CASE ALM_SEND
               /* Here, str1 is a linked list of strings, starting with list and sender,
                  followed by the entire mail (headers first). */
               examine_message (str1)
               DisposeLink (str1)

            ENDSELECT
         ENDIF
      ELSE     /* Deflag b */
         b := FALSE
      ENDIF
   ENDWHILE
   DisposeLink (buff)
ENDPROC


/*
 * Similar to barnacle_put_msg(), but not.  (:
 */
PROC boat_put_msg (reply, cmd, arg, field1, field2)
   DEF msg:PTR TO almsg, tmp

   NEW msg
   msg.mn.replyport  := alist_replyport
   msg.mn.length     := SIZEOF almsg
   msg.command       := cmd
   msg.arg           := arg
   msg.field1        := field1
   msg.field2        := field2

   /* Assume the port is still valid, because that's the protocol. */
   PutMsg (reply, msg)
   tmp := WaitPort (alist_replyport)
   Remove (tmp)
   END msg
ENDPROC


